import { Client, TextChannel, Message } from 'discord.js';
import log from 'npmlog';
import { BotConfig } from './types';
import { GoogleSpreadsheetWorksheet } from 'google-spreadsheet';
import { format } from 'string-kit';

export class Bot<Details> {
  config: BotConfig<Details>;
  discordClient: Client;
  globalChannelId?: string;
  constructor(config: BotConfig<Details>, globalChannelId?: string) {
    this.config = config;
    this.globalChannelId = globalChannelId;
    this.discordClient = new Client();

    log.info(this.config.name, 'instantiating bot...');

    this.discordClient.on('ready', () => {
      log.info(this.config.name, 'bot ready');
    });
  }
}

export async function useManualMessageSending(
  this: Bot<{
    channelIds: {
      dedicated: string;
      control: string;
    };
  }>,
  message: Message
): Promise<void> {
  if (message.channel.id !== this.config.details.channelIds.control) return;
  if (message.author.bot) return;
  log.info(
    this.config.name,
    'got message from %s in control channel',
    message.author.username
  );
  const channel = await this.discordClient.channels.fetch(
    this.config.details.channelIds.dedicated
  );
  if (channel instanceof TextChannel) {
    try {
      await channel.send(message.content);
      await message.react('✨');

      log.verbose(this.config.name, 'message successfully sent');
    } catch (e) {
      log.error(this.config.name, 'error sending or reacting to message', e);
    }
  } else {
    log.error(
      this.config.name,
      'could not send message because target channel is not a text channel'
    );
  }
}

export async function useClueChecking(
  this: Bot<{
    channelIds: { dedicated: string };
    guessHeader: string;
    responseHeader: string;
  }> & {
    stemsSheet: GoogleSpreadsheetWorksheet;
    phrasesSheet: GoogleSpreadsheetWorksheet;
  },
  message: Message,
  caseSensitive = false
): Promise<void> {
  if (message.channel.id !== this.config.details.channelIds.dedicated) return;
  if (message.author.bot) return;
  const guess = caseSensitive
    ? message.content.trim()
    : message.content.trim().toLocaleLowerCase();
  const stemsRows = await this.stemsSheet.getRows();
  const allPhrasesRows = await this.phrasesSheet.getRows();
  const positivePhrasesRows = allPhrasesRows.filter(
    row => row?.['positive']?.length
  );
  const negativePhrasesRows = allPhrasesRows.filter(
    row => row?.['negative']?.length
  );
  const match = stemsRows
    .map(row => row[this.config.details.guessHeader])
    .findIndex(
      answer =>
        guess ===
        (caseSensitive ? answer.trim() : answer.trim().toLocaleLowerCase())
    );
  log.verbose(
    this.config.name,
    guess
  );
  try {
    if (match !== -1) {
      const response = stemsRows[match]?.[this.config.details.responseHeader];

      if (response) {
        const positivePhrase =
          positivePhrasesRows[random(positivePhrasesRows.length)]?.['positive'];
        await message.channel.send(format(positivePhrase, response));
        log.info(
          this.config.name,
          'successful guess made, replied with clue: %j',
          response
        );
      } else {
        log.error(
          this.config.name,
          'response to valid guess (%j) is missing (%j)',
          guess,
          response
        );
      }
    } else {
      const negativePhrase =
        negativePhrasesRows[random(negativePhrasesRows.length)]?.['negative'];
      await message.channel.send(negativePhrase);
      log.info(
        this.config.name,
        'unsuccessful guess made, sent response: %j',
        negativePhrase
      );
    }
  } catch (e) {
    log.error(this.config.name, 'could not respond to message', e);
  }
}

export async function sendTimedClues(
  this: Bot<{ channelIds: { dedicated: string } }> & {
    stemsSheet: GoogleSpreadsheetWorksheet;
  },
  header: string,
  range: [number, number]
): Promise<void> {
  while (true) {
    await delay(random(range[0], range[1], false) * 60 * 1000);
    try {
      const allRows = await this.stemsSheet.getRows();
      const rows = allRows.filter(
        row => row[header]?.length && row['sent'] !== 'TRUE'
      );
      const channel = await this.discordClient.channels.fetch(
        this.config.details.channelIds.dedicated
      );
      if (channel instanceof TextChannel) {
        if (rows.length > 0) {
          const rowIndex = random(rows.length);
          const clue = rows[rowIndex]?.[header];
          if (!clue) {
            log.warn(this.config.name, 'clue not found: %j', clue);
          } else {
            await channel.send(clue);
            log.info(this.config.name, 'sent timed clue: %j', clue);
            const newRowIndex = rows[rowIndex].rowIndex;
            const newRow = allRows.find(row => row.rowIndex === newRowIndex);
            if (newRow) {
              newRow['sent'] = 'TRUE';
              await newRow.save();
              log.info(this.config.name, 'saved sent state for %j', clue);
            } else {
              log.warn(
                this.config.name,
                'sent state could not be saved for %j',
                clue
              );
            }
          }
        } else {
          log.warn(
            this.config.name,
            "no more clues available that haven't been sent"
          );
        }
      } else {
        log.error(
          this.config.name,
          'could not send message because target channel is not a text channel'
        );
      }
    } catch (e) {
      log.error(this.config.name, 'error sending message', e);
    }
  }
}

export async function sendTimedRandomMessages(
  this: Bot<unknown> & {
    phrasesSheet: GoogleSpreadsheetWorksheet;
  },
  range: [number, number],
  allowRandom = false
): Promise<void> {
  while (true) {
    await delay(random(range[0], range[1], false) * 60 * 1000);
    try {
      if (!this.globalChannelId) {
        throw new Error('Global channel missing from config');
      }
      const allRows = await this.phrasesSheet.getRows();
      const rows = allRows.filter(row => row['random']?.length);
      const channel = await this.discordClient.channels.fetch(
        this.globalChannelId
      );
      if (channel instanceof TextChannel) {
        let message = '';
        if (allowRandom && Math.random() > 0.8) {
          message = randomString();
        } else {
          message = rows[random(rows.length)]?.['random'];
        }
        if (message) {
          if (message.length > 2000) {
            const splitMessages = message.match(/[\s\S]{1,2000}/g);
            if (splitMessages) {
              await Promise.all(
                splitMessages.map(async (message: string) => {
                  await channel.send(message);
                })
              );
            } else {
              log.warn(
                this.config.name,
                'could not split message: %j (this should not happen)',
                message
              );
            }
          } else {
            await channel.send(message);
          }
          log.info(this.config.name, 'sent random message: %j', message);
        } else {
          log.warn(this.config.name, 'random message not found: %j', message);
        }
      } else {
        log.error(
          this.config.name,
          'could not send message because target channel is not a text channel'
        );
      }
    } catch (e) {
      log.error(this.config.name, 'error sending message', e);
    }
  }
}

export function random(a: number, b = 0, floor?: boolean): number {
  const min = Math.min(a, b);
  const max = Math.max(a, b);
  floor = floor ?? (Number.isInteger(a) && Number.isInteger(b));
  if (floor) {
    return Math.floor(Math.random() * (max - min) + min);
  } else {
    return Math.random() * (max - min) + min;
  }
}

function delay(ms: number) {
  return new Promise(res => setTimeout(res, ms));
}

export async function asyncFilter<T>(
  arr: T[],
  predicate: (arg: T) => Promise<boolean>
): Promise<T[]> {
  const results = await Promise.all(arr.map(predicate));

  return arr.filter((_v, index) => results[index]);
}

export function randomString(): string {
  let str = '';
  for (let i = 0; i <= random(1, 2000); i++) {
    str += String.fromCodePoint(random(6000));
  }
  log.info('random', 'using random string generator, generated %j', str);
  return str;
}
